/*
 * $Header: /opt/cvs/mumbletree/mumbletree.c,v 1.6 1999/10/22 08:44:29 morton Exp $
 *
 * $Log: mumbletree.c,v $
 * Revision 1.6  1999/10/22 08:44:29  morton
 * Added spillage optimisation
 *
 * Revision 1.5  1999/10/22 08:12:30  morton
 * Made test dir, rewrote mumbletree core
 *
 * Revision 1.4  1999/10/21 16:33:05  morton
 * *** empty log message ***
 *
 * Revision 1.3  1999/10/21 16:21:56  morton
 * Added left and right drop merging
 *
 * Revision 1.2  1999/10/21 14:43:29  morton
 * *** empty log message ***
 *
 * Revision 1.1.1.1  1999/10/21 14:42:04  morton
 * First cut
 *
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>

#include "mumbletree.h"

static void die(mt_node *_root);

static mt_node *mt_freelist = 0;

mt_node *
mt_new_node()
{
	mt_node *ret = mt_freelist;
	if (ret)
		mt_freelist = ret->left;
	else
		ret = malloc(sizeof(*ret));
	return ret;
}

void
mt_destroy_node(mt_node *_mt)
{
	_mt->left = mt_freelist;
	mt_freelist = _mt;
}

void
mt_destroy_tree(mt_node *_mt)
{
	if (_mt->left)
		mt_destroy_tree(_mt->left);
	if (_mt->right)
		mt_destroy_tree(_mt->right);
	mt_destroy_node(_mt);
}

mt_node *
mt_set_range(	mt_node *_root,
		const mt_value _low,
		const mt_value _high,
		const mt_payload _payload)
{
	assert(_low <= _high);

	if (_high < _root->low)
	{	/* It's to the left */
		/*
		 * FIXME: optimise for the case where _high == _root->left->high
		 * and the payloads is the same as root's
		 */
		mt_set_range(_root->left, _low, _high, _payload);

		/* Can we now merge left into root? */
		if (	_root->payload == _root->left->payload &&
			_root->left->high == _root->low - 1)
		{	/* Yes */
			mt_node *left = _root->left;
			assert(left->right == 0);
			_root->low = left->low;
			_root->left = left->left;
			mt_destroy_node(left);
		}
	}
	else if (_low > _root->high)
	{	/* It's to the right */
		mt_set_range(_root->right, _low, _high, _payload);

		/* Can we now merge right into root? */
		if (	_root->payload == _root->right->payload &&
			_root->right->low == _root->high + 1)
		{	/* Yes */
			mt_node *right = _root->right;
			assert(right->left == 0);
			_root->high = right->high;
			_root->right = right->right;
			mt_destroy_node(right);
		}
	}
	else
	{	/* It is within this node */
		if (_payload == _root->payload)
		{
			fprintf(stderr, "that's odd!\n");
		}

		if (_low == _root->low)
		{	/* It's left-aligned */
			if (_high == _root->high)
			{	/* It's right aligned as well */

				mt_node *left = _root->left;
				mt_node *right = _root->right;
				_root->payload = _payload;

				/* Can we merge in left? */
				if (left && left->payload == _payload)
				{	/* We can */
					assert(left->right == 0);
					_root->low = left->low;
					_root->left = left->left;
					mt_destroy_node(left);
				}
				/* Can we merge in right? */
				if (right && right->payload == _payload)
				{
					assert(right->left == 0);
					_root->high = right->high;
					_root->right = right->right;
					mt_destroy_node(right);
				}
			}
			else
			{	/* It's only left aligned */
				mt_node *left = _root->left;

				/* Can we spill data into left? */
				if (left && left->payload == _payload && left->high == _low - 1)
				{	/* Yes */
					left->high = _high;
					_root->low = _high + 1;
				}
				else
				{	/* No.  Break a new node off to the left */
					mt_node *new_node = mt_new_node();
					new_node->low = _low;
					new_node->high = _high;
					new_node->right = 0;
					new_node->left = _root->left;
					new_node->payload = _payload;
					_root->low = _high + 1;
					_root->left = new_node;
				}
			}
		}
		else if (_high == _root->high)
		{	/* It's only right aligned */
			mt_node *right = _root->right;

			/* Can we spill data into right? */
			if (right && right->payload == _payload && right->low == _high + 1)
			{	/* Yes */
				right->low = _low;
				_root->high = _low - 1;
			}
			else
			{	/* No.  Break off a new node to the right */
				mt_node *new_node = mt_new_node();
				new_node->low = _low;
				new_node->high = _high;
				new_node->left = 0;
				new_node->right = _root->right;
				new_node->payload = _payload;
				_root->high = _low - 1;
				_root->right = new_node;
			}
		}
		else
		{	/* It splits root */
			mt_node *new_left = mt_new_node();
			mt_node *new_right = mt_new_node();

			if (_low <= _root->low || _high >= _root->high)
			{
				fprintf(stderr,
					"bad overlap: [%d,%d) at node [%d,%d)\n",
					_low, _high,
					_root->low, _root->high);
				die(_root);
			}

			/*
			 * We create new left and right and modify *probe in place.
			 * Slightly less efficient, but it leaves the tree root at
			 * the same address, which is nice if it's the tree root
			 */
			new_left->low = _root->low;
			new_left->high = _low - 1;
			new_left->left = _root->left;
			new_left->right = 0;
			new_left->payload = _root->payload;

			new_right->low = _high + 1;
			new_right->high = _root->high;
			new_right->left = 0;
			new_right->right = _root->right;
			new_right->payload = _root->payload;

			_root->low = _low;
			_root->high = _high;
			_root->left = new_left;
			_root->right = new_right;
			_root->payload = _payload;
		}
	}

	return _root;
}

static int _mt_print(mt_node *_root, FILE *_file, int _depth, int _spaces)
{
	if (_depth)
	{
		char *mnem;

		switch (_root->payload)
		{
		case mt_e_assert:	mnem = "as"; break;
		case mt_e_nouse:	mnem = "nu"; break;
		case mt_e_readonly:	mnem = "ro"; break;
		case mt_e_readwrite:	mnem = "rw"; break;
		default: abort();
		}

		if (_root->right)
			_mt_print(_root->right, _file, _depth - 1, _spaces + 2);
		fprintf(_file, "% *s[%d,%d]->%s\n", _spaces, "", _root->low, _root->high, mnem);
		if (_root->left)
			_mt_print(_root->left, _file, _depth - 1, _spaces + 2);
	}
	return 0;
}

int mt_print(mt_node *_root, FILE *_file, int _depth, int _spaces)
{
	int ret;

	fprintf(_file, "Printing\n");
	ret = _mt_print(_root, _file, _depth, _spaces);
	fputc('\n', _file);
	return ret;
}

int
mt_good_range(	mt_node *_root,
		const mt_value _low,
		const mt_value _high,
		const mt_payload _payload)
{
	return 1;
}

/* Return next non-white; pointer points at it */
static char peek(char **_s)
{
	while (isspace(**_s))
		(*_s)++;
	return **_s;
}

/* Return next char */
static char wgrab(char **_s)
{
	char ret = **_s;
	if (ret)
		(*_s)++;
	return ret;
}

/* Return next non-white; pointer points beyond it */
static char grab(char **_s)
{
	char ret = peek(_s);
	if (ret)
		(*_s)++;
	return ret;
}

mt_payload grabp(char **_s)
{
	char c1 = grab(_s);
	char c2 = wgrab(_s);
	if (c1 == 'n' && c2 == 'u')
		return mt_e_nouse;
	if (c1 == 'r' && c2 == 'o')
		return mt_e_readonly;
	if (c1 == 'r' && c2 == 'w')
		return mt_e_readwrite;
	return mt_e_assert;
}

/* Return next integer */
static int grabi(char **_s)
{
	int ret = 0;
	while (isdigit(peek(_s)))
	{
		ret = ret * 10 + grab(_s) - '0';
	}
	return ret;
}

/*
 * Convert the passed string into a mumbletree.
 *
 * Format of expr:
 *
 * (low,high,payload,left,right)
 *
 * eg:
 *            10,20,ro
 *      1,9,rw       21,25,rw
 *
 * is
 *
 * (10,20,2,(1,9,3,0,0),(21,25,2,0,0))
 *
 * Return 0 in *success if things are OK.
 * Relies on the enumerated values of enum mt_payload.  Sorry.
 */

mt_node *
mt_string_to_tree(char **_expr, int *_success)
{
  mt_node *retval = 0;

  if (grab(_expr) == '(')
  {
    if (peek(_expr) == ')')
    {
       (*_expr)++;
    }
    else
    {
      int suc = 1;
      int low = grabi(_expr);
      if (grab(_expr) == ',')
      {
        int high = grabi(_expr);
        if (grab(_expr) == ',')
        {
          mt_payload payload = grabp(_expr);
          if (payload != mt_e_assert && grab(_expr) == ',')
          {
            mt_node *left = mt_string_to_tree(_expr, _success);
            if (*_success == 0 && grab(_expr) == ',')
            {
              mt_node *right = mt_string_to_tree(_expr, _success);
              if (*_success == 0 && grab(_expr) == ')')
              {
                 retval = mt_new_node();
                 retval->low = low;
                 retval->high = high;
                 retval->left = left;
                 retval->right = right;
                 retval->payload = payload;
                 suc = 0;
              }
            }
            else
            {
              if (left)
		mt_destroy_node(left);
            }
          }
        }
      }
      if (suc == 1)
	*_success = 1;
    }
  }

  return retval;
}

/*
 * Compare trees t1 amd t2.
 * Return < 0 if t1 < t2
 * Return 0 if t1 == t2
 * Return > 0 if t1 > t2
 */
int mt_compare_trees(mt_node *_t1, mt_node *_t2)
{
	int ret;

	if (_t1 == 0)
	{
		if (_t2 == 0)
			ret = 0;
		else
			ret = 1;
	}
	else if (_t2 == 0)
	{
		ret = -1;
	}
	else
	{
		if (	_t1->low < _t2->low ||
			_t1->high < _t2->high ||
			_t1->payload < _t2->payload)
		{
			return -1;
		}
		else if (	_t1->low > _t2->low ||
				_t1->high > _t2->high ||
				_t1->payload > _t2->payload)
		{
			return 1;
		}
		else if ((ret = mt_compare_trees(_t1->left, _t2->left)) == 0)
		{
			ret = mt_compare_trees(_t1->right, _t2->right);
		}
	}
	return ret;
}

static void die(mt_node *_root)
{
	mt_print(_root, stderr, -1, 0);
	abort();
}

